#!/usr/bin/env python3
# type: ignore
from __future__ import annotations

import ibis
import ibis.expr.datatypes as dt
import pytest
from pytest import param
from sqlalchemy import false
from sqlalchemy import true

SINGLESTOREDB_TYPES = [
    ('bool', dt.int8),
    ('boolean', dt.int8),
    ('bit', dt.int64),
    #   ("bit(1)", dt.int64),
    #   ("bit(9)", dt.int64),
    #   ("bit(17)", dt.int64),
    #   ("bit(33)", dt.int64),
    ('tinyint', dt.int8),
    ('tinyint unsigned', dt.uint8),
    ('smallint', dt.int16),
    ('smallint unsigned', dt.uint16),
    ('mediumint', dt.int32),
    ('mediumint unsigned', dt.uint32),
    ('int', dt.int32),
    ('int unsigned', dt.uint32),
    ('integer', dt.int32),
    ('integer unsigned', dt.uint32),
    ('bigint', dt.int64),
    ('bigint unsigned', dt.uint64),
    ('int1', dt.int8),
    ('int1 unsigned', dt.uint8),
    ('int2', dt.int16),
    ('int2 unsigned', dt.uint16),
    ('int3', dt.int32),
    ('int3 unsigned', dt.uint32),
    ('int4 unsigned', dt.uint32),
    ('float', dt.float32),
    ('double', dt.float64),
    ('decimal(10, 0)', dt.Decimal(10, 0)),
    ('decimal(5, 2)', dt.Decimal(5, 2)),
    ('dec', dt.Decimal(10, 0)),
    ('dec(5, 2)', dt.Decimal(5, 2)),
    ('numeric', dt.Decimal(10, 0)),
    ('fixed', dt.Decimal(10, 0)),
    ('date', dt.date),
    ('time', dt.time),
    ('time(6)', dt.time),
    ('datetime', dt.timestamp),
    ('datetime(6)', dt.timestamp),
    ('timestamp', dt.Timestamp('UTC')),
    ('timestamp(6)', dt.Timestamp('UTC')),
    ('year', dt.uint8),
    ('char(32)', dt.string),
    ('binary(42)', dt.binary),
    ('varchar(42)', dt.string),
    ('char byte', dt.binary),
    ('text', dt.string),
    ('varbinary(42)', dt.binary),
    ('longtext', dt.string),
    ('mediumtext', dt.string),
    ('text', dt.string),
    ('tinytext', dt.string),
    ('text(4)', dt.string),
    ('longblob', dt.binary),
    ('mediumblob', dt.binary),
    ('blob', dt.binary),
    ('tinyblob', dt.binary),
    ('json', dt.json),
    ('geographypoint', dt.string),
    ('geography', dt.string),
    ("enum('small', 'medium', 'large')", dt.string),
    ("set('a', 'b', 'c', 'd')", dt.string),
    #   ("bit(7)", dt.binary)
]


@pytest.mark.parametrize(
    ('singlestoredb_type', 'expected_type'),
    [
        param(singlestoredb_type, ibis_type, id=singlestoredb_type)
        for singlestoredb_type, ibis_type in SINGLESTOREDB_TYPES
    ],
)
def test_get_schema_from_query(con, singlestoredb_type, expected_type):
    if 'unsigned' in singlestoredb_type and \
            ('http://' in str(con.con.url) or 'https://' in str(con.con.url)):
        pytest.skip('HTTP API does not surface unsigned int information')
        return

    raw_name = ibis.util.guid()
    name = con.con.dialect.identifier_preparer.quote_identifier(raw_name)
    # temporary tables get cleaned up by the db when the session ends, so we
    # don't need to explicitly drop the table
    con.raw_sql(
        f'CREATE ROWSTORE TEMPORARY TABLE {name} '
        f'(x {singlestoredb_type}, y {singlestoredb_type})',
    )
    expected_schema = ibis.schema(dict(x=expected_type, y=expected_type))
    result_schema = con._get_schema_using_query(f'SELECT * FROM {name}')
    assert are_equal(
        result_schema, expected_schema,
    ), f'{result_schema} : {expected_schema}'

    if singlestoredb_type != 'geography':
        raw_name = ibis.util.guid()
        name = con.con.dialect.identifier_preparer.quote_identifier(raw_name)
        con.raw_sql(
            f'CREATE TEMPORARY TABLE {name} '
            f'(x {singlestoredb_type}, y {singlestoredb_type})',
        )
        expected_schema = ibis.schema(dict(x=expected_type, y=expected_type))
        result_schema = con._get_schema_using_query(f'SELECT * FROM {name}')
        assert are_equal(
            result_schema, expected_schema,
        ), f'{result_schema} : {expected_schema}'


def are_equal(result_schema, expected_schema):
    name_check = result_schema.names == expected_schema.names
    type_check = true

    if result_schema.types == (
        dt.Decimal(precision=None, scale=None, nullable=True),
        dt.Decimal(precision=None, scale=None, nullable=True),
    ) or \
            result_schema.types == (dt.Int64(nullable=True), dt.Int64(nullable=True)):

        for i, stype in enumerate(expected_schema.types):
            for name, expected_value in zip(stype.argnames, stype.args):
                result_value = getattr(result_schema.types[i], name, None)
                if result_value is not None and expected_value != result_value:
                    type_check = false
    else:
        type_check = result_schema.types == expected_schema.types

    return name_check and type_check



#!/usr/bin/env python
# type: ignore
"""SingleStoreDB connection tests."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Generator

import ibis
import pytest
from ibis.backends.tests.base import BackendTest
from ibis.backends.tests.base import RoundHalfToEven
from singlestoredb.connection import Connection
from singlestoredb.tests import utils

ibis.options.interactive = False
ibis.options.sql.default_limit = None
ibis.options.verbose = False


class TestConf(BackendTest, RoundHalfToEven):
    """
    SingleStoreDB connection tests.

    Parameters
    ----------
    data_directory : Path
        Path to input data

    """

    # singlestoredb has the same rounding behavior as postgres
    check_dtype = False
    supports_window_operations = False
    returned_timestamp_unit = 's'
    supports_arrays = False
    supports_arrays_outside_of_select = supports_arrays
    bool_is_int = True

    def __init__(self, data_directory: Path) -> None:
        super().__init__(data_directory)
        # mariadb supports window operations after version 10.2
        # but the sqlalchemy version string looks like:
        # 5.5.5.10.2.12.MariaDB.10.2.12+maria~jessie
        # or 10.4.12.MariaDB.1:10.4.12+maria~bionic
        # example of possible results:
        # https://github.com/sqlalchemy/sqlalchemy/blob/rel_1_3/
        # test/dialect/mysql/test_dialect.py#L244-L268
        self.__class__.supports_window_operations = True

    @staticmethod
    def connect(data_directory: Path) -> Connection:
        """
        Connect to SingleStoreDB database.

        Parameters
        ----------
        data_directory : Path
            Path to input data

        Returns
        -------
        Connection

        """
        return ibis.singlestoredb.connect()


def _random_identifier(suffix):
    return f'__ibis_test_{suffix}_{ibis.util.guid()}'


@pytest.fixture(scope='session')
def con():
    sql_file = os.path.join(os.path.dirname(__file__), 'test.sql')
    dbname, dbexisted = utils.load_sql(sql_file)

    yield ibis.singlestoredb.connect(database=dbname)

    if not dbexisted:
        utils.drop_database(dbname)


@pytest.fixture(scope='module')
def db(con):
    return con.database()


@pytest.fixture(scope='module')
def alltypes(db):
    return db.functional_alltypes


@pytest.fixture(scope='module')
def geotable(con):
    return con.table('geo')


@pytest.fixture(scope='module')
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope='module')
def gdf(geotable):
    return geotable.execute()


@pytest.fixture(scope='module')
def at(alltypes):
    return alltypes.op().sqla_table


@pytest.fixture(scope='module')
def intervals(con):
    return con.table('intervals')


@pytest.fixture
def temp_table(con) -> Generator[str, None, None]:
    """
    Return a temporary table name.

    Parameters
    ----------
    con : ibis.singlestoredb.SingleStoreDBClient

    Yields
    ------
    name : string
        Random table name for a temporary usage.

    """
    name = _random_identifier('table')
    try:
        yield name
    finally:
        con.drop_table(name, force=True)



#!/usr/bin/env python
"""SingleStoreDB function registry."""
from __future__ import annotations

from typing import Callable

import ibis
import ibis.backends.base.sql.compiler.translator as tr
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
import pandas as pd
import sqlalchemy as sa
from ibis.backends.base.sql.alchemy import fixed_arity
from ibis.backends.base.sql.alchemy import infix_op
from ibis.backends.base.sql.alchemy import reduction
from ibis.backends.base.sql.alchemy import sqlalchemy_operation_registry
from ibis.backends.base.sql.alchemy import sqlalchemy_window_functions_registry
from ibis.backends.base.sql.alchemy import unary
from ibis.backends.base.sql.alchemy import variance_reduction

operation_registry = sqlalchemy_operation_registry.copy()
operation_registry.update(sqlalchemy_window_functions_registry)


def _substr(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    f = sa.func.substr

    arg, start, length = expr.op().args

    sa_arg = t.translate(arg)
    sa_start = t.translate(start)

    if length is None:
        return f(sa_arg, sa_start + 1)
    else:
        sa_length = t.translate(length)
        return f(sa_arg, sa_start + 1, sa_length)


def _capitalize(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    (arg,) = expr.op().args
    sa_arg = t.translate(arg)
    return sa.func.concat(
        sa.func.ucase(sa.func.left(sa_arg, 1)), sa.func.substring(sa_arg, 2),
    )


def _extract(fmt: str) -> Callable[[tr.ExprTranslator, ir.Expr], ir.Expr]:
    def translator(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
        (arg,) = expr.op().args
        sa_arg = t.translate(arg)
        if fmt == 'millisecond':
            return sa.extract('microsecond', sa_arg) % 1000
        return sa.extract(fmt, sa_arg)

    return translator


_truncate_formats = {
    's': '%Y-%m-%d %H:%i:%s',
    'm': '%Y-%m-%d %H:%i:00',
    'h': '%Y-%m-%d %H:00:00',
    'D': '%Y-%m-%d',
    # 'W': 'week',
    'M': '%Y-%m-01',
    'Y': '%Y-01-01',
}


def _truncate(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    arg, unit = expr.op().args
    sa_arg = t.translate(arg)
    try:
        fmt = _truncate_formats[unit]
    except KeyError:
        raise com.UnsupportedOperationError(
            f'Unsupported truncate unit {unit}',
        )
    return sa.func.date_format(sa_arg, fmt)


def _cast(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    arg, typ = expr.op().args

    sa_arg = t.translate(arg)
    sa_type = t.get_sqla_type(typ)

    # specialize going from an integer type to a timestamp
    if isinstance(arg.type(), dt.Integer) and isinstance(sa_type, sa.DateTime):
        return sa.func.timezone('UTC', sa.func.to_timestamp(sa_arg))

    if arg.type().equals(dt.binary) and typ.equals(dt.string):
        return sa.func.hex(sa_arg)

    if typ.equals(dt.binary):
        #  decode yields a column of memoryview which is annoying to deal with
        # in pandas. CAST(expr AS BYTEA) is correct and returns byte strings.
        return sa.cast(sa_arg, sa.LargeBinary())

    return sa.cast(sa_arg, sa_type)


def _log(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    arg, base = expr.op().args
    sa_arg = t.translate(arg)
    sa_base = t.translate(base)
    return sa.func.log(sa_base, sa_arg)


def _identical_to(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    left, right = args = expr.op().args
    if left.equals(right):
        return True
    else:
        left, right = map(t.translate, args)
        return left.op('<=>')(right)


def _round(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    arg, digits = expr.op().args
    sa_arg = t.translate(arg)

    if digits is None:
        sa_digits = 0
    else:
        sa_digits = t.translate(digits)

    return sa.func.round(sa_arg, sa_digits)


def _floor_divide(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    left, right = map(t.translate, expr.op().args)
    return sa.func.floor(left / right)


def _string_join(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    sep, elements = expr.op().args
    return sa.func.concat_ws(t.translate(sep), *map(t.translate, elements))


def _interval_from_integer(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    arg, unit = expr.op().args
    if unit in {'ms', 'ns'}:
        raise com.UnsupportedOperationError(
            'SingleStoreDB does not allow operation '
            'with INTERVAL offset {}'.format(unit),
        )

    sa_arg = t.translate(arg)
    text_unit = expr.type().resolution.upper()

    # XXX: Is there a better way to handle this? I.e. can we somehow use
    # the existing bind parameter produced by translate and reuse its name in
    # the string passed to sa.text?
    if isinstance(sa_arg, sa.sql.elements.BindParameter):
        return sa.text(f'INTERVAL :arg {text_unit}').bindparams(
            arg=sa_arg.value,
        )
    return sa.text(f'INTERVAL {sa_arg} {text_unit}')


def _timestamp_diff(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    left, right = expr.op().args
    sa_left = t.translate(left)
    sa_right = t.translate(right)
    return sa.func.timestampdiff(sa.text('SECOND'), sa_right, sa_left)


def _literal(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    if isinstance(expr, ir.IntervalScalar):
        if expr.type().unit in {'ms', 'ns'}:
            raise com.UnsupportedOperationError(
                'SingleStoreDB does not allow operation '
                'with INTERVAL offset {}'.format(expr.type().unit),
            )
        text_unit = expr.type().resolution.upper()
        value = expr.op().value
        return sa.text(f'INTERVAL :value {text_unit}').bindparams(value=value)
    elif isinstance(expr, ir.SetScalar):
        return list(map(sa.literal, expr.op().value))
    else:
        value = expr.op().value
        if isinstance(value, pd.Timestamp):
            value = value.to_pydatetime()
        return sa.literal(value)


def _random(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    return sa.func.random()


def _group_concat(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    op = expr.op()
    arg, sep, where = op.args
    if where is not None:
        case = where.ifelse(arg, ibis.NA)
        arg = t.translate(case)
    else:
        arg = t.translate(arg)
    return sa.func.group_concat(arg.op('SEPARATOR')(t.translate(sep)))


def _day_of_week_index(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    (arg,) = expr.op().args
    left = sa.func.dayofweek(t.translate(arg)) - 2
    right = 7
    return ((left % right) + right) % right


def _day_of_week_name(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    (arg,) = expr.op().args
    return sa.func.dayname(t.translate(arg))


def _string_find(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    op = expr.op()

    if op.end is not None:
        raise NotImplementedError('`end` not yet implemented')

    if op.start is not None:
        return sa.func.locate(
            t.translate(op.substr),
            t.translate(op.arg),
            t.translate(op.start),
        ) - 1

    return sa.func.locate(t.translate(op.substr), t.translate(op.arg)) - 1


def _string_contains(t: tr.ExprTranslator, expr: ir.Expr) -> ir.Expr:
    return _string_find(t, expr) >= 0


operation_registry.update(
    {
        ops.Literal: _literal,
        # strings
        ops.Substring: _substr,
        ops.StringFind: _string_find,
        ops.StringContains: _string_contains,
        ops.Capitalize: _capitalize,
        ops.RegexSearch: infix_op('REGEXP'),
        ops.Cast: _cast,
        # math
        ops.Log: _log,
        ops.Log2: unary(sa.func.log2),
        ops.Log10: unary(sa.func.log10),
        ops.Round: _round,
        ops.RandomScalar: _random,
        # dates and times
        ops.Date: unary(sa.func.date),
        ops.DateAdd: infix_op('+'),
        ops.DateSub: infix_op('-'),
        ops.DateDiff: fixed_arity(sa.func.datediff, 2),
        ops.TimestampAdd: infix_op('+'),
        ops.TimestampSub: infix_op('-'),
        ops.TimestampDiff: _timestamp_diff,
        ops.DateTruncate: _truncate,
        ops.TimestampTruncate: _truncate,
        ops.IntervalFromInteger: _interval_from_integer,
        ops.Strftime: fixed_arity(sa.func.date_format, 2),
        ops.ExtractYear: _extract('year'),
        ops.ExtractMonth: _extract('month'),
        ops.ExtractDay: _extract('day'),
        ops.ExtractDayOfYear: unary('dayofyear'),
        ops.ExtractQuarter: _extract('quarter'),
        ops.ExtractEpochSeconds: unary('UNIX_TIMESTAMP'),
        ops.ExtractWeekOfYear: fixed_arity('weekofyear', 1),
        ops.ExtractHour: _extract('hour'),
        ops.ExtractMinute: _extract('minute'),
        ops.ExtractSecond: _extract('second'),
        ops.ExtractMillisecond: _extract('millisecond'),
        # reductions
        ops.BitAnd: reduction(sa.func.bit_and),
        ops.BitOr: reduction(sa.func.bit_or),
        ops.BitXor: reduction(sa.func.bit_xor),
        ops.Variance: variance_reduction('var'),
        ops.StandardDev: variance_reduction('stddev'),
        ops.IdenticalTo: _identical_to,
        ops.TimestampNow: fixed_arity(sa.func.now, 0),
        # others
        ops.GroupConcat: _group_concat,
        ops.DayOfWeekIndex: _day_of_week_index,
        ops.DayOfWeekName: _day_of_week_name,
        ops.HLLCardinality: reduction(
            lambda arg: sa.func.count(arg.distinct()),
        ),
    },
)



#!/usr/bin/env python3
from __future__ import annotations

from functools import partial
from typing import Any
from typing import Tuple

import ibis.expr.datatypes as dt

# binary character set
# used to distinguish blob binary vs blob text
MY_CHARSET_BIN = 63


def _type_from_cursor_info(descr: Tuple[Any, ...], field: Any) -> dt.DataType:
    """Construct an ibis type from MySQL field descr and field result metadata.

    This method is complex because the MySQL protocol is complex.

    Types are not encoded in a self contained way, meaning you need
    multiple pieces of information coming from the result set metadata to
    determine the most precise type for a field. Even then, the decoding is
    not high fidelity in some cases: UUIDs for example are decoded as
    strings, because the protocol does not appear to preserve the logical
    type, only the physical type.
    """
    _, type_code, _, _, field_length, scale, _ = descr
    flags = _FieldFlags(field.flags)
    typename = _type_codes.get(type_code)
    if typename is None:
        raise NotImplementedError(
            f'MySQL type code {type_code:d} is not supported',
        )

    typ = _type_mapping[typename]

    if typename in ('DECIMAL', 'NEWDECIMAL'):
        precision = _decimal_length_to_precision(
            length=field_length,
            scale=scale,
            is_unsigned=flags.is_unsigned,
        )
        typ = partial(typ, precision=precision, scale=scale)

    elif typename == 'BIT':
        typ = dt.int64

    elif typename == 'YEAR':
        typ = dt.uint8

    else:
        if flags.is_set:
            # sets are limited to strings
            typ = dt.Set(dt.string)
        elif flags.is_unsigned and type_code in _num_types:
            typ = getattr(dt, f'U{typ.__name__}')
        elif type_code in _char_types:
            # binary text
            if field.charsetnr == MY_CHARSET_BIN:
                typ = dt.Binary
            else:
                typ = dt.String

    # projection columns are always nullable
    return typ(nullable=True)


# ported from my_decimal.h:my_decimal_length_to_precision in mariadb
def _decimal_length_to_precision(
    *,
    length: int,
    scale: int,
    is_unsigned: bool,
) -> int:
    return length - (scale > 0) - (not (is_unsigned or not length))


_type_codes = {
    0: 'DECIMAL',
    1: 'TINY',
    2: 'SHORT',
    3: 'LONG',
    4: 'FLOAT',
    5: 'DOUBLE',
    6: 'NULL',
    7: 'TIMESTAMP',
    8: 'LONGLONG',
    9: 'INT24',
    10: 'DATE',
    11: 'TIME',
    12: 'DATETIME',
    13: 'YEAR',
    15: 'VARCHAR',
    16: 'BIT',
    245: 'JSON',
    246: 'NEWDECIMAL',
    247: 'ENUM',
    248: 'SET',
    249: 'TINY_BLOB',
    250: 'MEDIUM_BLOB',
    251: 'LONG_BLOB',
    252: 'BLOB',
    253: 'VAR_STRING',
    254: 'STRING',
    255: 'GEOMETRY',
}

_num_types = {1, 2, 3, 4, 5, 8, 9}
_char_types = {15, 249, 250, 251, 252, 253, 254, 255}

_type_mapping = {
    'DECIMAL': dt.Decimal,
    'TINY': dt.Int8,
    'SHORT': dt.Int16,
    'LONG': dt.Int32,
    'FLOAT': dt.Float32,
    'DOUBLE': dt.Float64,
    'NULL': dt.Null,
    'TIMESTAMP': lambda nullable: dt.Timestamp(
        timezone='UTC',
        nullable=nullable,
    ),
    'LONGLONG': dt.Int64,
    'INT24': dt.Int32,
    'DATE': dt.Date,
    'TIME': dt.Time,
    'DATETIME': dt.Timestamp,
    'YEAR': dt.Int16,
    'VARCHAR': dt.String,
    'BIT': dt.Int8,
    'JSON': dt.JSON,
    'NEWDECIMAL': dt.Decimal,
    'ENUM': dt.String,
    'SET': lambda nullable: dt.Set(dt.string, nullable=nullable),
    'TINY_BLOB': dt.Binary,
    'MEDIUM_BLOB': dt.Binary,
    'LONG_BLOB': dt.Binary,
    'BLOB': dt.Binary,
    'VAR_STRING': dt.String,
    'STRING': dt.String,
    'GEOMETRY': dt.Geometry,
}


class _FieldFlags:
    """Flags used to disambiguate field types.

    Gaps in the flag numbers are because we do not map in flags that are of no
    use in determining the field's type, such as whether the field is a primary
    key or not.
    """

    UNSIGNED = 1 << 5
    SET = 1 << 11
    NUM = 1 << 15

    __slots__ = ('value',)

    def __init__(self, value: int) -> None:
        self.value = value

    @property
    def is_unsigned(self) -> bool:
        return (self.UNSIGNED & self.value) != 0

    @property
    def is_set(self) -> bool:
        return (self.SET & self.value) != 0

    @property
    def is_num(self) -> bool:
        return (self.NUM & self.value) != 0



from __future__ import annotations

import ibis.expr.datatypes as dt
import sqlalchemy.dialects.mysql as singlestoredb
import toolz
from ibis.backends.base.sql.alchemy import AlchemyCompiler
from ibis.backends.base.sql.alchemy import AlchemyExprTranslator
from ibis.backends.base.sql.alchemy.registry import _geospatial_functions

from .registry import operation_registry


class SingleStoreDBExprTranslator(AlchemyExprTranslator):
    # https://dev.singlestore.com/doc/refman/8.0/en/spatial-function-reference.html
    _registry = toolz.merge(operation_registry, _geospatial_functions)
    _rewrites = AlchemyExprTranslator._rewrites.copy()
    _type_map = AlchemyExprTranslator._type_map.copy()
    _type_map.update(
        {
            dt.Boolean: singlestoredb.BOOLEAN,
            dt.Int8: singlestoredb.TINYINT,
            dt.Int16: singlestoredb.INTEGER,
            dt.Int32: singlestoredb.INTEGER,
            dt.Int64: singlestoredb.BIGINT,
            dt.Float16: singlestoredb.FLOAT,
            dt.Float32: singlestoredb.FLOAT,
            dt.Float64: singlestoredb.DOUBLE,
            dt.String: singlestoredb.VARCHAR,
        },
    )
    _bool_aggs_need_cast_to_int32 = False


rewrites = SingleStoreDBExprTranslator.rewrites


class SingleStoreDBCompiler(AlchemyCompiler):
    translator_class = SingleStoreDBExprTranslator



#!/usr/bin/env python3
from __future__ import annotations

import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import ibis
import ibis.backends.base.sql.compiler.translator as tr
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.rules as rlz
import ibis.expr.types as types
import sqlalchemy as sa


_db2py_type: Dict[str, str] = {
    'int64': 'int',
    'bigint': 'int',
    'string': 'str',
    'varchar': 'str',
    'text': 'str',
}


def _get_py_type(typ: Optional[str]) -> str:
    """
    Return the Python type for a given database type.

    Parameters
    ----------
    typ : str
        Name of the database type

    Returns
    -------
    str

    """
    if typ is None:
        return 'None'
    return _db2py_type.get(typ, typ)


def build_function(conn: Any, name: str) -> Callable[..., Any]:
    quote_identifier = conn.con.dialect.identifier_preparer.quote_identifier
    db = quote_identifier(conn.database_name)
    qname = quote_identifier(name)
    proto = conn.raw_sql(f'show create function {db}.{qname}').fetchall()[0][2]
    proto = re.split(r'\bfunction\s+', proto, flags=re.I)[-1]
    name, proto = proto.split('(', 1)

    if re.search(r'\)\s+returns\s+', proto, flags=re.I):
        sig, ret = re.split(r'\)\s+returns\s+', proto, flags=re.I)
        ret, ftype = re.split(r'\s+as\s+', ret, flags=re.I)
    else:
        ret = None
        sig, ftype = re.split(r'\s+as\s+', proto, flags=re.I)

    ftype, info = ftype.split(' ', 1)
    ftype = ftype.strip()

    m = re.search(r"^(.*)'\s+format\s+(\w+)\s*;\s*$", info, flags=re.I)
    if m is None:
        code = ''
        format = ''
    else:
        code = m.group(1)
        format = m.group(2)

    if name.startswith('`'):
        name = name[1:-1]

    input_names: List[str] = []
    inputs: List[str] = []
    for x in sig.split(','):
        m = re.match(r'^\s*(\w+)\s+(\w+)', x)
        if m is None:
            raise ValueError(f'Could not extract parameter names from: {sig}')
        input_names.append(m.group(1))
        inputs.append(m.group(2))

    nullable = [
        not re.search(r'\bnot\s+null\b', x, flags=re.I)
        for x in sig.split(',')
    ]

    inputs = [
        dict(
            bigint='int64', text='string', varchar='string',
            double='double',
        )[x] for x in inputs
    ]

    out_nullable = False
    output = ret
    if output:
        out_nullable = not re.search(r'\bnot\s+null\b', output, flags=re.I)
        m = re.match(r'^\s*(\w+)', output)
        if m is None:
            raise ValueError(f'Could not extract nullable information from: {output}')
        output = dict(
            bigint='int64', text='string',
            varchar='string', double='double',
        )[m.group(1)]

    return _make_udf(
        name, ftype.lower(),
        list(zip(input_names, inputs, nullable)),
        (output, out_nullable), code, format,
    )


def _make_func_doc(
    name: str,
    ftype: str,
    inputs: Sequence[Tuple[str, str, bool]],
    output: Optional[Tuple[str, bool]],
    code: str,
    format: str,
) -> str:
    """
    Construct the docstring using the function information.

    Parameters
    ----------
    name : str
        Name of the function
    ftype : str
        Type of the function in the database
    inputs : Sequence[Tuple[str, str, bool]]
        Sequence of (name, type, is_nullable) elements describing
        the inputs of the function
    output : Tuple[str, bool], optional
        Tuple of the form (type, is_nullable) for the return value
        of the function
    code : str
        Code of the UDF / UDA
    format : str
        UDF / UDA output format

    Returns
    -------
    str

    """
    doc = [f'Call `{name}` {ftype} function.', '']
    if ftype == 'remote service':
        doc.append(f'Accesses remote service at {code} using {format} format.')
        doc.append('')
    doc.extend(['Parameters', '----------'])
    for name, dtype, nullable in inputs:
        dtype = _get_py_type(dtype)
        arg = f'{name} : {dtype}'
        if nullable:
            arg += ' or None'
        doc.append(arg)
    if output and output[0]:
        doc.append('')
        doc.extend(['Returns', '-------'])
        ret = '{}'.format(_get_py_type(output[0]))
        if output[1]:
            ret += ' or None'
        doc.append(ret)
    doc.append('')
    return '\n'.join(doc)


def _make_udf(
    name: str,
    ftype: str,
    inputs: Sequence[Tuple[str, str, bool]],
    output: Optional[Tuple[str, bool]],
    code: str,
    format: str,
) -> Callable[..., Any]:
    """Define a callable that invokes a UDF on the server."""
    # Define generic function API
    cls_params = {}
    for arg, dtype, nullable in inputs:
        cls_params[arg] = getattr(rlz, dtype)
    if output:
        cls_params['output_dtype'] = getattr(dt, output[0])
        cls_params['output_shape'] = rlz.shape_like(inputs[0][0])
    func_type = type(name.title(), (ops.ValueOp,), cls_params)

    def eval_func(*args: Any) -> types.Expr:
        return func_type(*args).to_expr()

    eval_func.__name__ = name
    eval_func.__qualname__ = f'ibis.singlestoredb.{name}'
    eval_func.__doc__ = _make_func_doc(name, ftype, inputs, output, code, format)

    # TODO: Check for existing function
    if output:
        if output[0] == 'string':
            setattr(types.StringValue, name, eval_func)
        elif output[0] == 'int':
            setattr(types.IntegerValue, name, eval_func)

    @ibis.singlestoredb.add_operation(func_type)
    def _eval_func(t: tr.ExprTranslator, expr: types.Expr) -> types.Expr:
        return getattr(sa.func, name)(*[t.translate(x) for x in expr.op().args])

    return eval_func



"""The SingleStoreDB backend."""
from __future__ import annotations

import atexit
import contextlib
import warnings
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
import ibis.expr.types as ir
import pandas as pd
import sqlalchemy as sa
import sqlalchemy.dialects.mysql as singlestoredb
from ibis.backends.base.sql.alchemy import BaseAlchemyBackend
from ibis.backends.base.sql.registry.helpers import quote_identifier
from ibis.expr.types import AnyColumn
from pandas.io.json import build_table_schema
from singlestoredb.connection import build_params
from sqlalchemy_singlestoredb.base import SingleStoreDBDialect

from . import functions as fn
from .compiler import SingleStoreDBCompiler
from .datatypes import _type_from_cursor_info


__version__ = '0.1.0'


def _series_sqlalchemy_type(
    col: pd.Series,
    dtype: Optional[Dict[str, Any]] = None,
) -> Any:
    """
    Determine the SQLAlchemy type for a given pd.Series

    Parameters
    ----------
    col: pd.Series
        The pd.Series to inspect

    dtype: Dict[str, Any], optional
        Dictionary of data type overrides

    Returns
    -------
    SQLAlchemy data type

    """

    if dtype is not None and col.name in dtype:
        return dtype[col.name]

    # NOTE: It's dangerous to import private libraries, but we want to match
    #       their behavior as close as possible.
    import pandas._libs.lib as lib

    # Infer type of column, while ignoring missing values.
    # Needed for inserting typed data containing NULLs, GH 8778.
    col_type = lib.infer_dtype(col, skipna=True)

    import sqlalchemy.types as st

    if col_type == 'datetime64' or col_type == 'datetime':
        # GH 9086: TIMESTAMP is the suggested type if the column contains
        # timezone information
        try:
            if col.dt.tz is not None:
                return st.TIMESTAMP(timezone=True)
        except AttributeError:
            # The column is actually a DatetimeIndex
            # GH 26761 or an Index with date-like data e.g. 9999-01-01
            if getattr(col, 'tz', None) is not None:
                return st.TIMESTAMP(timezone=True)
        return st.DateTime

    if col_type == 'timedelta64':
        warnings.warn(
            "the 'timedelta' type is not supported, and will be "
            'written as integer values (ns frequency) to the database.',
            UserWarning,
            stacklevel=8,
        )
        return st.BigInteger

    elif col_type == 'floating':
        if col.dtype == 'float32':
            return st.Float(precision=23)
        else:
            return st.Float(precision=53)

    elif col_type == 'integer':
        # GH35076 Map pandas integer to optimal SQLAlchemy integer type
        if col.dtype.name.lower() in ('int8', 'uint8', 'int16'):
            return st.SmallInteger
        elif col.dtype.name.lower() in ('uint16', 'int32'):
            return st.Integer
        elif col.dtype.name.lower() == 'uint64':
            raise ValueError('Unsigned 64 bit integer datatype is not supported')
        else:
            return st.BigInteger

    elif col_type == 'boolean':
        return st.Boolean

    elif col_type == 'date':
        return st.Date

    elif col_type == 'time':
        return st.Time

    elif col_type == 'complex':
        raise ValueError('Complex datatypes not supported')

    elif col_type == 'decimal':
        return st.DECIMAL(60, 30)

    return st.Text


def _ibis_schema_to_sqlalchemy_dtypes(df_schema: ibis.Schema) -> Dict[str, Any]:
    """
    Convert an Ibis Schema to a dict of SQLAlchemy types.

    Parameters
    ----------
    schema: ibis.Schema
        Schema object to convert

    Returns
    -------
    Dict[str, Any]

    """
    from ibis.backends.base.sql.alchemy import datatypes
    return dict(
        zip(
            df_schema.names,
            [datatypes.to_sqla_type(x) for x in df_schema.types],
        ),
    )


def _infer_dtypes(
    frame: pd.DataFrame,
    dtype: Optional[Dict[str, Any]] = None,
) -> Dict[str, Any]:
    """
    Infer the SQLAlchemy dtypes for a DataFrame.

    Parameters
    ----------
    frame : pd.DataFrame
        The DataFrame to inspect

    Returns
    -------
    Dict[str, sa.type]

    """
    return dict([
        (str(frame.columns[i]), _series_sqlalchemy_type(frame.iloc[:, i], dtype))
        for i in range(len(frame.columns))
    ])


class Backend(BaseAlchemyBackend):
    name = 'singlestoredb'
    compiler = SingleStoreDBCompiler

    _database_name: Optional[str] = None

    @property
    def database_name(self) -> Optional[str]:
        """Get the currently selected database."""
        return self._database_name

    @database_name.setter
    def database_name(self, value: Optional[str]) -> None:
        """Set the default database name."""
        # TODO: unset database
        if value is None:
            return

        # TODO: escape value
        value = str(value)
        if self._database_name != value and hasattr(self, 'con'):
            self.raw_sql(f'use {value}')

        self._database_name = value

    def create_database(self, name: str, force: bool = False) -> None:
        """
        Create a new database.
        Parameters
        ----------
        name : str
            Name for the new database
        force : bool, optional
            If `True`, an exception is raised if the database already exists.
        """
        if force and name.lower() in [x.lower() for x in self.list_databases()]:
            raise ValueError(f'Database with the name "{name}" already exists.')
        # TODO: escape name
        self.raw_sql(f'CREATE DATABASE IF NOT EXISTS {name}')

    def sync_functions(self) -> None:
        """Synchronize client APIs with server functions."""
        for row in self.raw_sql('SHOW FUNCTIONS'):
            fn.build_function(self, row[0])

    def do_connect(self, *args: str, **kwargs: Any) -> None:
        """Connect to a SingleStoreDB database."""
        if args:
            params = build_params(host=args[0], **kwargs)
        else:
            params = build_params(**kwargs)

        driver = params.pop('driver', None)
        if driver and not driver.startswith('singlestoredb+'):
            driver = 'singlestoredb+{}'.format(driver)

        alchemy_url = self._build_alchemy_url(
            url=params.pop('url', None),
            host=params.pop('host', None),
            port=params.pop('port', None),
            user=params.pop('user', None),
            password=params.pop('password', None),
            database=params.pop('database', None),
            driver=driver,
        )

        alchemy_url.set(query={k: str(v) for k, v in params.items()})

        self.database_name = alchemy_url.database

        super().do_connect(
            sa.create_engine(
                alchemy_url,
                echo=kwargs.get('echo', False), future=kwargs.get('future', False),
            ),
        )

        self.sync_functions()

#   @contextlib.contextmanager
#   def begin(self) -> Generator[Any, Any, Any]:
#       with super().begin() as bind:
#           previous_timezone = bind.execute(
#               'SELECT @@session.time_zone',
#           ).scalar()
#           unset_timezone = False
#           try:
#               bind.execute("SET @@session.time_zone = 'UTC'")
#               unset_timezone = True
#           except Exception as e:
#               warnings.warn(f"Couldn't set singlestore timezone: {str(e)}")

#           try:
#               yield bind
#           finally:
#               if unset_timezone:
#                   query = "SET @@session.time_zone = '{}'"
#                   bind.execute(query.format(previous_timezone))

    def create_table(
        self,
        name: str,
        expr: pd.DataFrame | ir.TableExpr | None = None,
        schema: sch.Schema | None = None,
        database: str | None = None,
        force: bool = False,
    ) -> None:
        """Create a new table.
        Parameters
        ----------
        name
            Name of the new table.
        expr
            An Ibis table expression or pandas DataFrame that will be used to
            extract the schema and the data of the new table. If not provided,
            `schema` must be given.
        schema
            The schema for the new table. Only one of `schema` or `expr` can be
            provided.
        database
            Name of the database where the table will be created, if not the
            default.
        force
            Check whether a table exists before creating it
        """
        if database == self.current_database:
            # avoid fully qualified name
            database = None

        if database is not None:
            raise NotImplementedError(
                'Creating tables from a different database is not yet '
                'implemented',
            )

        if expr is None and schema is None:
            raise ValueError('You must pass either an expression or a schema')

        if isinstance(expr, pd.DataFrame):
            if schema is not None:
                pd_schema_names = ibis.pandas.connect(
                    {name: expr},
                ).table(name).schema().names
                if not sorted(pd_schema_names) == sorted(sch.schema(schema).names):
                    raise TypeError(
                        'Expression schema is not equal to passed schema. '
                        'Try passing the expression without the schema',
                    )

            # TODO: Should this be done in `insert` as well?
            expr = expr.copy()
            for column in expr:
                try:
                    expr[column].dt.tz_localize('UTC')
                except (AttributeError, TypeError):
                    pass

            if schema is not None:
                dtype = _ibis_schema_to_sqlalchemy_dtypes(schema)
            else:
                dtype = _infer_dtypes(expr)

            expr.to_sql(
                name,
                self.con,
                index=False,
                if_exists='replace' if force else 'fail',
                dtype=dtype,
            )

        elif isinstance(expr, ir.TableExpr) or schema is not None:
            if expr is not None and schema is not None:
                if not sorted(expr.schema().names) == sorted(sch.schema(schema).names):
                    raise TypeError(
                        'Expression schema is not equal to passed schema. '
                        'Try passing the expression without the schema',
                    )

            if schema is None:
                schema = expr.schema()  # type: ignore

            self._schemas[self._fully_qualified_name(name, database)] = schema
            t = self._table_from_schema(
                name, schema, database=database or self.current_database,
            )

            with self.begin() as bind:
                t.create(bind=bind, checkfirst=force)
                if expr is not None:
                    bind.execute(
                        t.insert().from_select(list(expr.columns), expr.compile()),
                    )

        else:
            raise TypeError(
                '`expr` and/or `schema` are not an expected type: {} / {}'.format(
                    type(expr).__name__, type(schema).__name__,
                ),
            )

        return self.table(name)

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        """Infer the schema of `query`."""
        result = self.con.execute(f'SELECT * FROM ({query}) _ LIMIT 0')
        cursor = result.cursor._cursor
        fields = [
            (field.name, _type_from_cursor_info(descr, field))
            for descr, field in zip(cursor.description, cursor._result.fields)
        ]
        return sch.Schema.from_tuples(fields)

    def _get_temp_view_definition(
        self,
        name: str,
        definition: sa.sql.compiler.Compiled,
    ) -> str:
        return f'CREATE OR REPLACE VIEW {name} AS {definition}'

    def _register_temp_view_cleanup(self, name: str, raw_name: str) -> None:
        query = f'DROP VIEW IF EXISTS {name}'

        def drop(self: Backend, raw_name: str, query: str) -> None:
            self.con.execute(query)
            self._temp_views.discard(raw_name)

        atexit.register(drop, self, raw_name, query)


# TODO(kszucs): unsigned integers


@dt.dtype.register((singlestoredb.DOUBLE, singlestoredb.REAL))
def singlestoredb_double(satype: Any, nullable: bool = True) -> dt.Float64:
    return dt.Float64(nullable=nullable)


@dt.dtype.register(singlestoredb.FLOAT)
def singlestoredb_float(satype: Any, nullable: bool = True) -> dt.Float32:
    return dt.Float32(nullable=nullable)


@dt.dtype.register(singlestoredb.TINYINT)
def singlestoredb_tinyint(satype: Any, nullable: bool = True) -> dt.Int8:
    return dt.Int8(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.YEAR)
def singlestoredb_year(dialect: Any, satype: Any, nullable: bool = True) -> dt.Int16:
    return dt.Int16(nullable=nullable)


@dt.dtype.register(singlestoredb.BLOB)
def singlestoredb_blob(satype: Any, nullable: bool = True) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.BIT)
def singlestoredb_bit(dialect: Any, satype: Any, nullable: bool = True) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.BINARY)
def singlestoredb_binary(dialect: Any, satype: Any, nullable: bool = True) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.VARBINARY)
def singlestoredb_varbinary(
    dialect: Any,
    satype: Any,
    nullable: bool = True,
) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.LONGBLOB)
def singlestoredb_longblob(dialect: Any, satype: Any, nullable: bool = True) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.MEDIUMBLOB)
def singlestoredb_mediumblob(
    dialect: Any,
    satype: Any,
    nullable: bool = True,
) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.TINYBLOB)
def singlestoredb_tinyblob(dialect: Any, satype: Any, nullable: bool = True) -> dt.Binary:
    return dt.Binary(nullable=nullable)


@dt.dtype.register(SingleStoreDBDialect, singlestoredb.JSON)
def singlestoredb_json(dialect: Any, satype: Any, nullable: bool = True) -> dt.JSON:
    return dt.JSON(nullable=nullable)
